from haversine import inverse_haversine_vector, Unit, Direction
from numpy import isclose
from math import pi
import pytest

from tests.geo_ressources import LYON, PARIS, NEW_YORK, LONDON


@pytest.mark.parametrize(
    "point, dir, dist, result",
    [
        (PARIS, Direction.NORTH, 32, (49.144444, 2.3508)),
        (PARIS, 0, 32, (49.144444, 2.3508)),
        (LONDON, Direction.WEST, 50, (51.507778, -0.840556)),
        (LONDON, pi * 1.5, 50, (51.507778, -0.840556)),
        (NEW_YORK, Direction.SOUTH, 15, (40.568611, -74.235278)),
        (NEW_YORK, Direction.NORTHWEST, 50, (41.020556, -74.656667)),
        (NEW_YORK, pi * 1.25, 50, (40.384722, -74.6525)),
    ],
)
def test_inverse_kilometers(point, dir, dist, result):
    assert isclose(inverse_haversine_vector([point], [dist], [dir]),
                   ([result[0]], [result[1]]), rtol=1e-5).all()

def test_inverse_normalization():
    twoDegreesAtEquator = 222390.1
    point = (0.0, -179.0)

    # non-breaking behavior without normalization
    result = inverse_haversine_vector(
        [point, LONDON], [twoDegreesAtEquator, 50], [Direction.WEST, Direction.WEST], Unit.METERS)

    # assert result is a Tuple
    assert isinstance(result, tuple)

    # assert result is a Tuple of array
    latArray, lngArray = result

    assert isclose(lngArray[0], -181.0, rtol=1e-5,)

    # behavior with normalization
    result = inverse_haversine_vector(
        [point, LONDON], [twoDegreesAtEquator, 50000], direction=[Direction.WEST, Direction.WEST],
        unit=Unit.METERS, normalize_output=True,)

    
    # assert result is a Tuple
    assert isinstance(result, tuple)

    # assert result is a Tuple of array
    latArray, lngArray = result

    assert isclose(lngArray[0], 179.0, rtol=1e-5,).all()
    assert isclose(lngArray[1], -0.840556, rtol=1e-5,).all()


from enum import Enum
from math import pi
from typing import Union, Tuple
import math


# mean earth radius - https://en.wikipedia.org/wiki/Earth_radius#Mean_radius
_AVG_EARTH_RADIUS_KM = 6371.0088


class Unit(str, Enum):
    """
    Enumeration of supported units.
    The full list can be checked by iterating over the class; e.g.
    the expression `tuple(Unit)`.
    """

    KILOMETERS = 'km'
    METERS = 'm'
    MILES = 'mi'
    NAUTICAL_MILES = 'nmi'
    FEET = 'ft'
    INCHES = 'in'
    RADIANS = 'rad'
    DEGREES = 'deg'


class Direction(float, Enum):
    """
    Enumeration of supported directions.
    The full list can be checked by iterating over the class; e.g.
    the expression `tuple(Direction)`.
    Angles expressed in radians.
    """

    NORTH = 0.0
    NORTHEAST = pi * 0.25
    EAST = pi * 0.5
    SOUTHEAST = pi * 0.75
    SOUTH = pi
    SOUTHWEST = pi * 1.25
    WEST = pi * 1.5
    NORTHWEST = pi * 1.75


# Unit values taken from http://www.unitconversion.org/unit_converter/length.html
_CONVERSIONS = {
    Unit.KILOMETERS:       1.0,
    Unit.METERS:           1000.0,
    Unit.MILES:            0.621371192,
    Unit.NAUTICAL_MILES:   0.539956803,
    Unit.FEET:             3280.839895013,
    Unit.INCHES:           39370.078740158,
    Unit.RADIANS:          1/_AVG_EARTH_RADIUS_KM,
    Unit.DEGREES:          (1/_AVG_EARTH_RADIUS_KM)*(180.0/pi)
}


def get_avg_earth_radius(unit):
    return _AVG_EARTH_RADIUS_KM * _CONVERSIONS[unit]


def _normalize(lat: float, lon: float) -> Tuple[float, float]:
    """
    Normalize point to [-90, 90] latitude and [-180, 180] longitude.
    """
    lat = (lat + 90) % 360 - 90
    if lat > 90:
        lat = 180 - lat
        lon += 180
    lon = (lon + 180) % 360 - 180
    return lat, lon


def _normalize_vector(lat: "numpy.ndarray", lon: "numpy.ndarray") -> Tuple["numpy.ndarray", "numpy.ndarray"]:
    """
    Normalize points to [-90, 90] latitude and [-180, 180] longitude.
    """
    lat = (lat + 90) % 360 - 90
    lon = (lon + 180) % 360 - 180
    wrap = lat > 90
    if numpy.any(wrap):
        lat[wrap] = 180 - lat[wrap]
        lon[wrap] = lon[wrap] % 360 - 180
    return lat, lon


def _ensure_lat_lon(lat: float, lon: float):
    """
    Ensure that the given latitude and longitude have proper values. An exception is raised if they are not.
    """
    if lat < -90 or lat > 90:
        raise ValueError(f"Latitude {lat} is out of range [-90, 90]")
    if lon < -180 or lon > 180:
        raise ValueError(f"Longitude {lon} is out of range [-180, 180]")


def _ensure_lat_lon_vector(lat: "numpy.ndarray", lon: "numpy.ndarray"):
    """
    Ensure that the given latitude and longitude have proper values. An exception is raised if they are not.
    """
    if numpy.abs(lat).max() > 90:
        raise ValueError("Latitude(s) out of range [-90, 90]")
    if numpy.abs(lon).max() > 180:
        raise ValueError("Longitude(s) out of range [-180, 180]")


def _explode_args(f):
    return lambda ops: f(**ops.__dict__)


@_explode_args
def _create_haversine_kernel(*, asin=None, arcsin=None, cos, radians, sin, sqrt, **_):
    asin = asin or arcsin

    def _haversine_kernel(lat1, lng1, lat2, lng2):
        """
        Compute the haversine distance on unit sphere.  Inputs are in degrees,
        either scalars (with ops==math) or arrays (with ops==numpy).
        """
        lat1 = radians(lat1)
        lng1 = radians(lng1)
        lat2 = radians(lat2)
        lng2 = radians(lng2)
        lat = lat2 - lat1
        lng = lng2 - lng1
        d = (sin(lat * 0.5) ** 2
             + cos(lat1) * cos(lat2) * sin(lng * 0.5) ** 2)
        # Note: 2 * atan2(sqrt(d), sqrt(1-d)) is more accurate at
        # large distance (d is close to 1), but also slower.
        return 2 * asin(sqrt(d))
    return _haversine_kernel



@_explode_args
def _create_inverse_haversine_kernel(*, asin=None, arcsin=None, atan2=None, arctan2=None, cos, degrees, radians, sin, sqrt, **_):
    asin = asin or arcsin
    atan2 = atan2 or arctan2

    def _inverse_haversine_kernel(lat, lng, direction, d):
        """
        Compute the inverse haversine on unit sphere.  lat/lng are in degrees,
        direction in radians; all inputs are either scalars (with ops==math) or
        arrays (with ops==numpy).
        """
        lat = radians(lat)
        lng = radians(lng)
        cos_d, sin_d = cos(d), sin(d)
        cos_lat, sin_lat = cos(lat), sin(lat)
        sin_d_cos_lat = sin_d * cos_lat
        return_lat = asin(cos_d * sin_lat + sin_d_cos_lat * cos(direction))
        return_lng = lng + atan2(sin(direction) * sin_d_cos_lat,
                                 cos_d - sin_lat * sin(return_lat))
        return degrees(return_lat), degrees(return_lng)
    return _inverse_haversine_kernel


_haversine_kernel = _create_haversine_kernel(math)
_inverse_haversine_kernel = _create_inverse_haversine_kernel(math)

try:
    import numpy
    has_numpy = True
    _haversine_kernel_vector = _create_haversine_kernel(numpy)
    _inverse_haversine_kernel_vector = _create_inverse_haversine_kernel(numpy)
except ModuleNotFoundError:
    # Import error will be reported in haversine_vector() / inverse_haversine_vector()
    has_numpy = False

try:
    import numba # type: ignore
    if has_numpy:
        _haversine_kernel_vector = numba.vectorize(fastmath=True)(_haversine_kernel_vector)
        # Tuple output is not supported for numba.vectorize. Just jit the numpy version.
        _inverse_haversine_kernel_vector = numba.njit(fastmath=True)(_inverse_haversine_kernel_vector)
    _haversine_kernel = numba.njit(_haversine_kernel)
    _inverse_haversine_kernel = numba.njit(_inverse_haversine_kernel)
except ModuleNotFoundError:
    pass


def haversine(point1, point2, unit=Unit.KILOMETERS, normalize=False, check=True):
    """ Calculate the great-circle distance between two points on the Earth surface.

    Takes two 2-tuples, containing the latitude and longitude of each point in decimal degrees,
    and, optionally, a unit of length.

    :param point1: first point; tuple of (latitude, longitude) in decimal degrees
    :param point2: second point; tuple of (latitude, longitude) in decimal degrees
    :param unit: a member of haversine.Unit, or, equivalently, a string containing the
                 initials of its corresponding unit of measurement (i.e. miles = mi)
                 default 'km' (kilometers).
    :param normalize: if True, normalize the points to [-90, 90] latitude and [-180, 180] longitude.
    :param check: if True, check that points are normalized.

    Example: ``haversine((45.7597, 4.8422), (48.8567, 2.3508), unit=Unit.METERS)``

    Precondition: ``unit`` is a supported unit (supported units are listed in the `Unit` enum)

    :return: the distance between the two points in the requested unit, as a float.

    The default returned unit is kilometers. The default unit can be changed by
    setting the unit parameter to a member of ``haversine.Unit``
    (e.g. ``haversine.Unit.INCHES``), or, equivalently, to a string containing the
    corresponding abbreviation (e.g. 'in'). All available units can be found in the ``Unit`` enum.
    """

    # unpack latitude/longitude
    lat1, lng1 = point1
    lat2, lng2 = point2

    # normalize points or ensure they are proper lat/lon, i.e., in [-90, 90] and [-180, 180]
    if normalize:
        lat1, lng1 = _normalize(lat1, lng1)
        lat2, lng2 = _normalize(lat2, lng2)
    elif check:
        _ensure_lat_lon(lat1, lng1)
        _ensure_lat_lon(lat2, lng2)

    return get_avg_earth_radius(unit) * _haversine_kernel(lat1, lng1, lat2, lng2)


def haversine_vector(array1, array2, unit=Unit.KILOMETERS, comb=False, normalize=False, check=True):
    '''
    The exact same function as "haversine", except that this
    version replaces math functions with numpy functions.
    This may make it slightly slower for computing the haversine
    distance between two points, but is much faster for computing
    the distance between two vectors of points due to vectorization.
    '''
    if not has_numpy:
        raise RuntimeError('Error, unable to import Numpy, '
                           'consider using haversine instead of haversine_vector.')

    # ensure arrays are numpy ndarrays
    if not isinstance(array1, numpy.ndarray):
        array1 = numpy.array(array1)
    if not isinstance(array2, numpy.ndarray):
        array2 = numpy.array(array2)

    # ensure will be able to iterate over rows by adding dimension if needed
    if array1.ndim == 1:
        array1 = numpy.expand_dims(array1, 0)
    if array2.ndim == 1:
        array2 = numpy.expand_dims(array2, 0)

    # Asserts that both arrays have same dimensions if not in combination mode
    if not comb:
        if array1.shape != array2.shape:
            raise IndexError(
                "When not in combination mode, arrays must be of same size. If mode is required, use comb=True as argument.")

    # unpack latitude/longitude
    lat1, lng1 = array1[:, 0], array1[:, 1]
    lat2, lng2 = array2[:, 0], array2[:, 1]

    # normalize points or ensure they are proper lat/lon, i.e., in [-90, 90] and [-180, 180]
    if normalize:
        lat1, lng1 = _normalize_vector(lat1, lng1)
        lat2, lng2 = _normalize_vector(lat2, lng2)
    elif check:
        _ensure_lat_lon_vector(lat1, lng1)
        _ensure_lat_lon_vector(lat2, lng2)

    # If in combination mode, turn coordinates of array1 into column vectors for broadcasting
    if comb:
        lat1 = numpy.expand_dims(lat1, axis=0)
        lng1 = numpy.expand_dims(lng1, axis=0)
        lat2 = numpy.expand_dims(lat2, axis=1)
        lng2 = numpy.expand_dims(lng2, axis=1)

    return get_avg_earth_radius(unit) * _haversine_kernel_vector(lat1, lng1, lat2, lng2)


def inverse_haversine(point, distance, direction: Union[Direction, float], unit=Unit.KILOMETERS, normalize_output=False):
    lat, lng = point
    r = get_avg_earth_radius(unit)
    outLat, outLng = _inverse_haversine_kernel(lat, lng, direction, distance / r)

    if normalize_output:
        return _normalize(outLat, outLng)
    else:
        return (outLat, outLng)


def inverse_haversine_vector(array, distance, direction, unit=Unit.KILOMETERS, normalize_output=False): # -> Tuple["numpy.ndarray", "numpy.ndarray"]:
    if not has_numpy:
        raise RuntimeError('Error, unable to import Numpy, '
                           'consider using inverse_haversine instead of inverse_haversine_vector.')

    # ensure arrays are numpy ndarrays
    array, distance, direction = map(numpy.asarray, (array, distance, direction))

    # ensure will be able to iterate over rows by adding dimension if needed
    if array.ndim == 1:
        array = numpy.expand_dims(array, 0)

    # Asserts that arrays are correctly sized
    if array.ndim != 2 or array.shape[1] != 2 or array.shape[0] != len(distance) or array.shape[0] != len(direction):
        raise IndexError("Arrays must be of same size.")

    # unpack latitude/longitude
    lat, lng = array[:, 0], array[:, 1]

    r = get_avg_earth_radius(unit)
    outLatArray, outLngArray = _inverse_haversine_kernel_vector(lat, lng, direction, distance/r)

    if normalize_output:
        return _normalize_vector(outLatArray, outLngArray)

    return (outLatArray, outLngArray)


from haversine import haversine_vector, Unit
from numpy.testing import assert_allclose
import pytest

from tests.geo_ressources import EXPECTED_LONDON_PARIS, EXPECTED_LYON_NEW_YORK, EXPECTED_LYON_PARIS, EXPECTED_LONDON_NEW_YORK, LYON, PARIS, NEW_YORK, LONDON


@pytest.mark.parametrize(
    'unit', [Unit.KILOMETERS, Unit.METERS, Unit.INCHES]
)
def test_pair(unit):
    def test_lyon_paris(unit):
        expected_lyon_paris = EXPECTED_LYON_PARIS[unit]
        assert haversine_vector(LYON, PARIS, unit=unit) == expected_lyon_paris
        assert isinstance(unit.value, str)
        assert haversine_vector(
            LYON, PARIS, unit=unit.value) == expected_lyon_paris

    return test_lyon_paris(unit)


@pytest.mark.parametrize(
    "oob_from,oob_to,proper_from,proper_to", [
        ((-90.0001, 0), (0, 0), (-89.9999, 180), (0, 0)),
        ((-90.0001, 30), (0, 0), (-89.9999, -150), (0, 0)),
        ((0, 0), (90.0001, 0), (0, 0), (89.9999, -180)),
        ((0, 0), (90.0001, 30), (0, 0), (89.9999, -150)),
        ((0, -180.0001), (0, 0), (0, 179.9999), (0, 0)),
        ((30, -180.0001), (0, 0), (30, 179.9999), (0, 0)),
        ((0, 0), (0, 180.0001), (0, 0), (0, -179.9999)),
        ((0, 0), (30, 180.0001), (0, 0), (30, -179.9999)),
    ]
)
def test_normalization(oob_from, oob_to, proper_from, proper_to):
    """
    Test makes sure that normalization works as expected by comparing distance of out of
    bounds points cases to equal cases where all points are within lat/lon ranges. The
    results are expected to be equal (within some tolerance to account for numerical
    issues).
    """
    normalized_during, normalized_already = (
        haversine_vector([oob_from], [oob_to], Unit.DEGREES, normalize=True),
        haversine_vector([proper_from], [proper_to],
                         Unit.DEGREES, normalize=True),
    )
    assert normalized_during == pytest.approx(normalized_already, abs=1e-10)


@pytest.mark.parametrize(
    "oob_from,oob_to", [
        ((-90.0001, 0), (0, 0)),
        ((0, 0), (90.0001, 0)),
        ((0, -180.0001), (0, 0)),
        ((0, 0), (0, 180.0001)),
    ]
)
def test_out_of_bounds(oob_from, oob_to):
    """
    Test makes sure that a ValueError is raised when latitude or longitude values are out of bounds.
    """
    with pytest.raises(ValueError):
        haversine_vector([oob_from], [oob_to])
    with pytest.raises(ValueError):
        haversine_vector([oob_from], [oob_to], normalize=False)


def test_haversine_vector_comb():
    unit = Unit.KILOMETERS
    expected = [
        [EXPECTED_LYON_PARIS[unit],  EXPECTED_LONDON_PARIS[unit]],
        [EXPECTED_LYON_NEW_YORK[unit], EXPECTED_LONDON_NEW_YORK[unit]]
    ]

    assert_allclose(  # See https://numpy.org/doc/stable/reference/generated/numpy.testing.assert_allclose.html#numpy.testing.assert_allclose
        haversine_vector([LYON, LONDON], [PARIS, NEW_YORK], unit, comb=True),
        expected
    )


def test_units_enum():
    from haversine.haversine import _CONVERSIONS
    assert all(unit in _CONVERSIONS for unit in Unit)



from haversine import haversine, Unit
from math import pi
import pytest

from tests.geo_ressources import LYON, PARIS, NEW_YORK, LONDON, EXPECTED_LYON_PARIS


def haversine_test_factory(unit):
    def test():
        expected = EXPECTED_LYON_PARIS[unit]
        assert haversine(LYON, PARIS, unit=unit) == expected
        assert isinstance(unit.value, str)
        assert haversine(LYON, PARIS, unit=unit.value) == expected

    return test


test_kilometers = haversine_test_factory(Unit.KILOMETERS)
test_meters = haversine_test_factory(Unit.METERS)
test_miles = haversine_test_factory(Unit.MILES)
test_nautical_miles = haversine_test_factory(Unit.NAUTICAL_MILES)
test_feet = haversine_test_factory(Unit.FEET)
test_inches = haversine_test_factory(Unit.INCHES)
test_radians = haversine_test_factory(Unit.RADIANS)
test_degrees = haversine_test_factory(Unit.DEGREES)


def test_units_enum():
    from haversine.haversine import _CONVERSIONS
    assert all(unit in _CONVERSIONS for unit in Unit)


def test_haversine_deg_rad():
    """
    Test makes sure that one time around earth matches sphere circumference in degrees / radians.
    """
    p1, p2 = (45, 0), (-45, 180)
    assert haversine(p1, p2, unit=Unit.RADIANS) == pi
    assert round(haversine(p1, p2, unit=Unit.DEGREES), 13) == 180.0


@pytest.mark.parametrize(
    "oob_from,oob_to,proper_from,proper_to", [
        ((-90.0001, 0), (0, 0), (-89.9999, 180), (0, 0)),
        ((-90.0001, 30), (0, 0), (-89.9999, -150), (0, 0)),
        ((0, 0), (90.0001, 0), (0, 0), (89.9999, -180)),
        ((0, 0), (90.0001, 30), (0, 0), (89.9999, -150)),
        ((0, -180.0001), (0, 0), (0, 179.9999), (0, 0)),
        ((30, -180.0001), (0, 0), (30, 179.9999), (0, 0)),
        ((0, 0), (0, 180.0001), (0, 0), (0, -179.9999)),
        ((0, 0), (30, 180.0001), (0, 0), (30, -179.9999)),
    ]
)
def test_normalization(oob_from, oob_to, proper_from, proper_to):
    """
    Test makes sure that normalization works as expected by comparing distance of out of
    bounds points cases to equal cases where all points are within lat/lon ranges. The
    results are expected to be equal (within some tolerance to account for numerical
    issues).
    """
    normalized_during, normalized_already = (
        haversine(oob_from, oob_to, Unit.DEGREES, normalize=True),
        haversine(proper_from, proper_to, Unit.DEGREES, normalize=True),
    )
    assert normalized_during == pytest.approx(normalized_already, abs=1e-10)


@pytest.mark.parametrize(
    "oob_from,oob_to", [
        ((-90.0001, 0), (0, 0)),
        ((0, 0), (90.0001, 0)),
        ((0, -180.0001), (0, 0)),
        ((0, 0), (0, 180.0001)),
    ]
)
def test_out_of_bounds(oob_from, oob_to):
    """
    Test makes sure that a ValueError is raised when latitude or longitude values are out of bounds.
    """
    with pytest.raises(ValueError):
        haversine(oob_from, oob_to)
    with pytest.raises(ValueError):
        haversine(oob_from, oob_to, normalize=False)


@pytest.mark.parametrize(
    "in_bounds_from,in_bounds_to", [
        ((-90, 0), (0, 0)),
        ((0, 0), (90, 0)),
        ((0, -180), (0, 0)),
        ((0, 0), (0, 180)),
    ]
)
def test_in_bounds(in_bounds_from, in_bounds_to):
    """
    Test makes sure that a ValueError is NOT raised when latitude or longitude values are in bounds.
    """
    assert haversine(in_bounds_from, in_bounds_to) > 0


def test_haversine_deg_rad_great_circle_distance():
    """
    Test makes sure the haversine functions returns the great circle distance (https://en.wikipedia.org/wiki/Great-circle_distance) between two points on a sphere.
    See https://github.com/mapado/haversine/issues/45
    """
    p1, p2 = (0, -45), (0, 45)
    assert haversine(p1, p2, Unit.DEGREES) == 89.99999999999997



from .  import haversine_baseline as baseline
import haversine as current
import numpy as np
import pytest
from timeit import timeit

def assert_performance(func, number):
    # Interleave measurements and compare fastest current to median baseline.
    # All in an attempt to avoid spurious errors caused by fluctuating load on
    # the runner.
    t_baseline, t_current = [], []
    for repeat in range(5):
        t_baseline.append(timeit(lambda: func(baseline),  number=number))
        t_current.append(timeit(lambda: func(current),  number=number))

    perf_ratio = np.min(t_current) / np.median(t_baseline)
    assert perf_ratio <= 1.1


@pytest.mark.parametrize(
    "check", [False, True]
)
@pytest.mark.parametrize(
    "normalize", [False, True]
)
def test_haversine(check, normalize):
    if check == normalize == True:
        pytest.skip()
    assert_performance(lambda m: m.haversine((0,1), (2,3), check=check, normalize=normalize),
                       number=100000)


@pytest.mark.parametrize(
    "check", [False, True]
)
@pytest.mark.parametrize(
    "normalize", [False, True]
)
def test_haversine_vector(check, normalize):
    if check == normalize == True:
        pytest.skip()
    arr = np.random.uniform(size=(1000, 2))
    assert_performance(lambda m: m.haversine_vector(arr, arr, check=check, normalize=normalize),
                       number=1000)


def test_inverse_haversine():
    assert_performance(lambda m: m.inverse_haversine((0,1), 2, 3),
                       number=100000)


def test_inverse_haversine_vector():
    arr = np.random.uniform(size=(1000, 2))
    assert_performance(lambda m: m.inverse_haversine_vector(arr, *arr.T),
                       number=1000)



from setuptools import setup

setup(
    name='haversine',
    version='2.9.0',
    description='Calculate the distance between 2 points on Earth.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    python_requires='>=3.5',
    author='Balthazar Rouberol',
    maintainer='Julien Deniau',
    maintainer_email='julien.deniau@mapado.com',
    url='https://github.com/mapado/haversine',
    packages=['haversine'],
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: GIS'
    ],
)



from haversine import inverse_haversine, haversine, Unit, Direction
from numpy import isclose
from math import pi
import pytest

from tests.geo_ressources import LYON, PARIS, NEW_YORK, LONDON


@pytest.mark.parametrize(
    "point, dir, dist, result",
    [
        (PARIS, Direction.NORTH, 32, (49.144444, 2.3508)),
        (PARIS, 0, 32, (49.144444, 2.3508)),
        (LONDON, Direction.WEST, 50, (51.507778, -0.840556)),
        (LONDON, pi * 1.5, 50, (51.507778, -0.840556)),
        (NEW_YORK, Direction.SOUTH, 15, (40.568611, -74.235278)),
        (NEW_YORK, Direction.NORTHWEST, 50, (41.020556, -74.656667)),
        (NEW_YORK, pi * 1.25, 50, (40.384722, -74.6525)),
    ],
)
def test_inverse_kilometers(point, dir, dist, result):
    assert isclose(inverse_haversine(point, dist, dir),
                   result, rtol=1e-5).all()


@pytest.mark.parametrize(
    "point, direction, distance, unit",
    [
        (PARIS, Direction.NORTH, 10, Unit.KILOMETERS),
        (LONDON, Direction.WEST, 32, Unit.MILES),
        (LYON, Direction.NORTHEAST, 45_000, Unit.METERS),
        (NEW_YORK, Direction.SOUTH, 15, Unit.NAUTICAL_MILES),
    ],
)
def test_back_and_forth(point, direction, distance, unit):
    new_point = inverse_haversine(point, distance, direction, unit)
    assert isclose(haversine(new_point, point, unit), distance, rtol=1e-10)


def test_inverse_miles():
    assert isclose(inverse_haversine(PARIS, 50, Direction.NORTH,
                   unit=Unit.MILES), (49.5803579218996, 2.3508), rtol=1e-5).all()


def test_nautical_inverse_miles():
    assert isclose(inverse_haversine(PARIS, 10, Direction.SOUTH,
                   unit=Unit.NAUTICAL_MILES), (48.69014586638915, 2.3508), rtol=1e-5).all()

def test_inverse_normalization():
    twoDegreesAtEquator = 222390.1
    # non-breaking behavior without normalization
    _, pointWestLon = inverse_haversine(
        (0.0, -179.0), twoDegreesAtEquator, Direction.WEST, Unit.METERS)
    assert isclose(pointWestLon, -181.0, rtol=1e-5,)

    # behavior with normalization
    _, pointWestLon = inverse_haversine(
        (0.0, -179.0), twoDegreesAtEquator, direction=Direction.WEST,
        unit=Unit.METERS, normalize_output=True,)
    assert isclose(pointWestLon, 179.0, rtol=1e-5,)

